/*
 * Copyright (c) 2022 Arm Limited.
 *
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#if defined(ARM_COMPUTE_ENABLE_SME2)

#include <cstddef>
#include <cstdint>

namespace arm_conv {
namespace depthwise {

void sme2_fp32_nhwc_3x3_s1_output4x4_mla_depthfirst_indirect_impl(
  const float *const *const input_ptrs,
  float *const *const outptrs,
  const void *params,
  unsigned int n_channels,
  const float activation_min,
  const float activation_max
)
{
  struct Args
  {
    float *const *outptrs;
    const void *params;
    const float min, max;
    const float *inptrs[36];

    Args(
      const float *const *const input_ptrs,
      float *const *const outptrs,
      const void *const params,
      const float min,
      const float max
    ) : outptrs(outptrs), params(params), min(min), max(max)
    {
      inptrs[0] = input_ptrs[14];
      inptrs[1] = input_ptrs[0];
      inptrs[2] = input_ptrs[5];
      inptrs[3] = input_ptrs[15];
      inptrs[4] = input_ptrs[30];
      inptrs[5] = input_ptrs[35];
      inptrs[6] = input_ptrs[20];
      inptrs[7] = input_ptrs[1];
      inptrs[8] = input_ptrs[4];
      inptrs[9] = input_ptrs[21];
      inptrs[10] = input_ptrs[6];
      inptrs[11] = input_ptrs[11];
      inptrs[12] = input_ptrs[24];
      inptrs[13] = input_ptrs[8];
      inptrs[14] = input_ptrs[29];
      inptrs[15] = input_ptrs[9];
      inptrs[16] = input_ptrs[31];
      inptrs[17] = input_ptrs[13];
      inptrs[18] = input_ptrs[34];
      inptrs[19] = input_ptrs[16];
      inptrs[20] = input_ptrs[2];
      inptrs[21] = input_ptrs[19];
      inptrs[22] = input_ptrs[3];
      inptrs[23] = input_ptrs[12];
      inptrs[24] = input_ptrs[22];
      inptrs[25] = input_ptrs[17];
      inptrs[26] = input_ptrs[18];
      inptrs[27] = input_ptrs[26];
      inptrs[28] = input_ptrs[23];
      inptrs[29] = input_ptrs[32];
      inptrs[30] = input_ptrs[27];
      inptrs[31] = input_ptrs[33];
      inptrs[32] = input_ptrs[7];
      inptrs[33] = input_ptrs[10];
      inptrs[34] = input_ptrs[25];
      inptrs[35] = input_ptrs[28];

    }
  };

  Args params_struct(input_ptrs, outptrs, params,
                     activation_min, activation_max);

  __asm__ __volatile__(
    "ldr x16, [%x[params_struct], %[offsetof_args_params]]\n"
    ".inst 0xd503477f  // SMSTART ZA\n"
    "add x15, %x[params_struct], %[offsetof_Args_inptrs]\n"
    "ptrue p3.b\n"
    ".inst 0x25207810  // ptrue pn8.b\n"
    "ld1w { z15.s }, p3/Z, [x16]\n"
    "addvl x16, x16, #1\n"
    "ldp x14, x13, [x15, #0x0]\n"
    "ldp x12, x11, [x15, #0x10]\n"
    "cntw x10\n"
    ".inst 0xa040c200  // ld1w { z0.s-z3.s }, pn8.b/Z, [x16]\n"
    "addvl x16, x16, #4\n"
    "mov x9, #0x0\n"
    "whilelt p2.s, XZR, %x[n_channels]\n"
    ".inst 0xa040c204  // ld1w { z4.s-z7.s }, pn8.b/Z, [x16]\n"
    "ldr x28, [%x[params_struct], %[offsetof_args_outptrs]]\n"
    "addvl x16, x16, #4\n"
    "cmp x10, %x[n_channels]\n"
    "ld1rw { z14.s }, p3/Z, [%x[params_struct], %[offsetof_args_min]]\n"
    "ld1rw { z13.s }, p3/Z, [%x[params_struct], %[offsetof_args_max]]\n"
    "sub x27, XZR, x10\n"
    "ld1w { z8.s }, p3/Z, [x16]\n"
    "addvl x16, x16, #1\n"
    "ld1w { z9.s }, p2/Z, [x14, x9, LSL #2]\n"
    "ld1w { z10.s }, p2/Z, [x13, x9, LSL #2]\n"
    "ld1w { z11.s }, p2/Z, [x12, x9, LSL #2]\n"
    "ld1w { z12.s }, p2/Z, [x11, x9, LSL #2]\n"
    "bge 2f\n"
    "1:"  // Channel loop
    "movprfx z21, z15\n fmla z21.s, p3/M, z4.s, z9.s\n"
    "movprfx z16, z15\n fmla z16.s, p3/M, z8.s, z9.s\n"
    "ldr x26, [x15, #0x20]\n"
    "incw x27\n"
    "movprfx z22, z15\n fmla z22.s, p3/M, z3.s, z9.s\n"
    "movprfx z25, z15\n fmla z25.s, p3/M, z1.s, z9.s\n"
    "ldr x25, [x15, #0x30]\n"
    "mov p1.b, p2.b\n"
    "movprfx z26, z15\n fmla z26.s, p3/M, z0.s, z9.s\n"
    "ldr x24, [x15, #0x28]\n"
    "movprfx z17, z15\n fmla z17.s, p3/M, z7.s, z9.s\n"
    "whilelt p0.s, x10, %x[n_channels]\n"
    "movprfx z18, z15\n fmla z18.s, p3/M, z6.s, z9.s\n"
    "fmla z21.s, p3/M, z5.s, z12.s\n"
    "ldr x23, [x15, #0x38]\n"
    "movprfx z20, z15\n fmla z20.s, p3/M, z5.s, z9.s\n"
    "movprfx z24, z15\n fmla z24.s, p3/M, z2.s, z9.s\n"
    "ld1w { z9.s }, p2/Z, [x25, x9, LSL #2]\n"
    "ldr x14, [x15, #0x40]\n"
    "fmla z16.s, p3/M, z0.s, z10.s\n"
    "movprfx z19, z15\n fmla z19.s, p3/M, z2.s, z11.s\n"
    "ld1w { z10.s }, p2/Z, [x26, x9, LSL #2]\n"
    "ldr x13, [x15, #0x48]\n"
    "fmla z22.s, p3/M, z4.s, z12.s\n"
    "fmla z25.s, p3/M, z2.s, z12.s\n"
    "ld1w { z11.s }, p2/Z, [x24, x9, LSL #2]\n"
    "ldr x12, [x15, #0x50]\n"
    "fmla z26.s, p3/M, z1.s, z12.s\n"
    "fmla z17.s, p3/M, z8.s, z12.s\n"
    "ldr x26, [x15, #0x60]\n"
    "fmla z18.s, p3/M, z7.s, z12.s\n"
    "movprfx z28, z15\n fmla z28.s, p3/M, z6.s, z10.s\n"
    "ld1w { z10.s }, p2/Z, [x13, x9, LSL #2]\n"
    "ldr x24, [x15, #0x68]\n"
    "fmla z21.s, p3/M, z7.s, z9.s\n"
    "fmla z19.s, p3/M, z6.s, z12.s\n"
    "ldr x11, [x15, #0x58]\n"
    "movprfx z23, z15\n fmla z23.s, p3/M, z3.s, z12.s\n"
    "movprfx z27, z15\n fmla z27.s, p3/M, z0.s, z12.s\n"
    "ld1w { z12.s }, p2/Z, [x23, x9, LSL #2]\n"
    "ldr x25, [x15, #0x70]\n"
    "movprfx z31, z15\n fmla z31.s, p3/M, z8.s, z11.s\n"
    "fmla z22.s, p3/M, z6.s, z9.s\n"
    "ld1w { z11.s }, p2/Z, [x14, x9, LSL #2]\n"
    "ldr x23, [x15, #0x78]\n"
    "fmla z25.s, p3/M, z4.s, z9.s\n"
    "fmla z26.s, p3/M, z3.s, z9.s\n"
    "ldr x14, [x15, #0x80]\n"
    "movprfx z29, z15\n fmla z29.s, p3/M, z1.s, z9.s\n"
    "movprfx z30, z15\n fmla z30.s, p3/M, z0.s, z9.s\n"
    "ldr x13, [x15, #0x88]\n"
    "ld1w { z15.s }, p3/Z, [x16]\n"
    "fmla z20.s, p3/M, z8.s, z9.s\n"
    "fmla z24.s, p3/M, z5.s, z9.s\n"
    "ldr x22, [x28, #0x0]\n"
    "addvl x16, x16, #1\n"
    "fmla z28.s, p3/M, z2.s, z9.s\n"
    "fmla z16.s, p3/M, z1.s, z12.s\n"
    "ld1w { z9.s }, p2/Z, [x12, x9, LSL #2]\n"
    "ldr x12, [x15, #0x90]\n"
    "fmla z17.s, p3/M, z0.s, z12.s\n"
    "fmla z18.s, p3/M, z2.s, z11.s\n"
    "ld1w { z12.s }, p2/Z, [x11, x9, LSL #2]\n"
    "ldr x11, [x15, #0x98]\n"
    "fmla z21.s, p3/M, z8.s, z10.s\n"
    "fmla z19.s, p3/M, z1.s, z11.s\n"
    "ld1w { z11.s }, p2/Z, [x26, x9, LSL #2]\n"
    "ldr x26, [x15, #0xa0]\n"
    "fmla z22.s, p3/M, z7.s, z10.s\n"
    "fmla z23.s, p3/M, z6.s, z10.s\n"
    "ldr x21, [x28, #0x8]\n"
    "fmla z25.s, p3/M, z5.s, z10.s\n"
    "fmla z26.s, p3/M, z4.s, z10.s\n"
    "ldr x20, [x28, #0x10]\n"
    "fmla z27.s, p3/M, z3.s, z10.s\n"
    "fmla z29.s, p3/M, z2.s, z10.s\n"
    "ldr x19, [x28, #0x18]\n"
    "fmla z30.s, p3/M, z1.s, z10.s\n"
    "fmla z31.s, p3/M, z0.s, z10.s\n"
    "ld1w { z10.s }, p2/Z, [x24, x9, LSL #2]\n"
    "ldr x24, [x15, #0xa8]\n"
    "fmla z16.s, p3/M, z3.s, z9.s\n"
    "fmla z20.s, p3/M, z0.s, z9.s\n"
    "fmla z24.s, p3/M, z6.s, z11.s\n"
    "fmla z28.s, p3/M, z3.s, z11.s\n"
    "ld1w { z11.s }, p2/Z, [x25, x9, LSL #2]\n"
    "ldr x25, [x15, #0xb0]\n"
    "fmla z17.s, p3/M, z4.s, z10.s\n"
    "fmla z18.s, p3/M, z3.s, z10.s\n"
    "fmla z21.s, p3/M, z1.s, z10.s\n"
    "fmla z19.s, p3/M, z5.s, z12.s\n"
    "fmla z23.s, p3/M, z2.s, z12.s\n"
    "fmla z22.s, p3/M, z0.s, z10.s\n"
    "ld1w { z12.s }, p2/Z, [x23, x9, LSL #2]\n"
    "ldr x23, [x15, #0xb8]\n"
    "fmla z27.s, p3/M, z8.s, z11.s\n"
    "fmla z31.s, p3/M, z5.s, z11.s\n"
    "ld1w { z11.s }, p2/Z, [x14, x9, LSL #2]\n"
    "ldr x14, [x15, #0xc0]\n"
    "fmla z16.s, p3/M, z5.s, z10.s\n"
    "fmla z20.s, p3/M, z2.s, z10.s\n"
    "ld1w { z10.s }, p2/Z, [x13, x9, LSL #2]\n"
    "ldr x13, [x15, #0xc8]\n"
    "fmla z17.s, p3/M, z5.s, z12.s\n"
    "fmla z18.s, p3/M, z4.s, z12.s\n"
    "fmla z21.s, p3/M, z2.s, z12.s\n"
    "fmla z19.s, p3/M, z3.s, z12.s\n"
    "fmla z22.s, p3/M, z1.s, z12.s\n"
    "fmla z23.s, p3/M, z0.s, z12.s\n"
    "ld1w { z12.s }, p2/Z, [x11, x9, LSL #2]\n"
    "ldr x11, [x15, #0xd8]\n"
    "fmla z28.s, p3/M, z7.s, z11.s\n"
    "fmla z29.s, p3/M, z6.s, z11.s\n"
    "ld1w { z11.s }, p2/Z, [x12, x9, LSL #2]\n"
    "ldr x12, [x15, #0xd0]\n"
    "fmla z16.s, p3/M, z7.s, z10.s\n"
    "fmla z17.s, p3/M, z6.s, z10.s\n"
    "fmla z20.s, p3/M, z4.s, z10.s\n"
    "fmla z21.s, p3/M, z3.s, z10.s\n"
    "fmla z24.s, p3/M, z1.s, z10.s\n"
    "fmla z25.s, p3/M, z0.s, z10.s\n"
    "ld1w { z10.s }, p2/Z, [x26, x9, LSL #2]\n"
    "ldr x26, [x15, #0xe0]\n"
    "fmla z18.s, p3/M, z8.s, z12.s\n"
    "fmla z30.s, p3/M, z8.s, z11.s\n"
    "fmla z31.s, p3/M, z7.s, z11.s\n"
    "ld1w { z11.s }, p2/Z, [x24, x9, LSL #2]\n"
    "fmla z27.s, p3/M, z1.s, z12.s\n"
    "ldr x24, [x15, #0xe8]\n"
    "fmla z19.s, p3/M, z7.s, z12.s\n"
    "fmla z22.s, p3/M, z5.s, z12.s\n"
    "fmla z23.s, p3/M, z4.s, z12.s\n"
    "fmla z26.s, p3/M, z2.s, z12.s\n"
    "ld1w { z12.s }, p2/Z, [x25, x9, LSL #2]\n"
    "ldr x25, [x15, #0xf0]\n"
    "fmla z16.s, p3/M, z2.s, z10.s\n"
    "fmla z17.s, p3/M, z1.s, z10.s\n"
    "fmla z18.s, p3/M, z0.s, z10.s\n"
    "fmla z20.s, p3/M, z7.s, z11.s\n"
    "ld1w { z10.s }, p2/Z, [x23, x9, LSL #2]\n"
    "ldr x23, [x15, #0xf8]\n"
    "fmla z21.s, p3/M, z6.s, z11.s\n"
    "fmla z24.s, p3/M, z4.s, z11.s\n"
    "fmla z25.s, p3/M, z3.s, z11.s\n"
    "fmla z28.s, p3/M, z1.s, z11.s\n"
    "fmla z29.s, p3/M, z0.s, z11.s\n"
    "ld1w { z11.s }, p2/Z, [x14, x9, LSL #2]\n"
    "fmla z27.s, p3/M, z4.s, z11.s\n"
    "ldr x14, [x15, #0x100]\n"
    "fmla z30.s, p3/M, z2.s, z11.s\n"
    "fmla z17.s, p3/M, z2.s, z12.s\n"
    "fmla z18.s, p3/M, z1.s, z12.s\n"
    "fmla z19.s, p3/M, z0.s, z12.s\n"
    "ld1w { z12.s }, p2/Z, [x13, x9, LSL #2]\n"
    "ldr x13, [x15, #0x108]\n"
    "fmla z16.s, p3/M, z6.s, z10.s\n"
    "fmla z20.s, p3/M, z3.s, z10.s\n"
    "fmla z24.s, p3/M, z0.s, z10.s\n"
    "fmla z22.s, p3/M, z8.s, z11.s\n"
    "ld1w { z10.s }, p2/Z, [x12, x9, LSL #2]\n"
    "ldr x12, [x15, #0x110]\n"
    "fmla z23.s, p3/M, z7.s, z11.s\n"
    "fmla z26.s, p3/M, z5.s, z11.s\n"
    "fmla z31.s, p3/M, z1.s, z11.s\n"
    "ld1w { z11.s }, p2/Z, [x11, x9, LSL #2]\n"
    "fmla z27.s, p3/M, z2.s, z12.s\n"
    "ldr x11, [x15, #0x118]\n"
    "fmla z28.s, p3/M, z0.s, z10.s\n"
    "fmla z29.s, p3/M, z4.s, z11.s\n"
    "fmla z30.s, p3/M, z3.s, z11.s\n"
    "fmla z19.s, p3/M, z8.s, z12.s\n"
    "fmla z23.s, p3/M, z5.s, z12.s\n"
    "fmla z20.s, p3/M, z6.s, z10.s\n"
    "ld1w { z12.s }, p2/Z, [x26, x9, LSL #2]\n"
    "fmla z24.s, p3/M, z3.s, z10.s\n"
    "ld1w { z10.s }, p2/Z, [x24, x9, LSL #2]\n"
    "fmla z25.s, p3/M, z7.s, z11.s\n"
    "fmla z26.s, p3/M, z6.s, z11.s\n"
    "fmla z28.s, p3/M, z5.s, z11.s\n"
    "fmla z27.s, p3/M, z5.s, z12.s\n"
    "fmla z31.s, p3/M, z2.s, z12.s\n"
    "fmla z29.s, p3/M, z7.s, z10.s\n"
    "fmla z30.s, p3/M, z6.s, z10.s\n"
    "fmla z24.s, p3/M, z8.s, z11.s\n"
    "ld1w { z11.s }, p2/Z, [x25, x9, LSL #2]\n"
    "fmla z28.s, p3/M, z8.s, z10.s\n"
    "fmla z25.s, p3/M, z8.s, z11.s\n"
    "fmla z26.s, p3/M, z7.s, z11.s\n"
    "ld1w { z10.s }, p2/Z, [x14, x9, LSL #2]\n"
    "fmla z27.s, p3/M, z6.s, z11.s\n"
    "fmla z29.s, p3/M, z5.s, z11.s\n"
    "fmla z30.s, p3/M, z4.s, z11.s\n"
    "fmla z31.s, p3/M, z3.s, z11.s\n"
    "ld1w { z11.s }, p2/Z, [x13, x9, LSL #2]\n"
    "ldp x14, x13, [x15, #0x0]\n"
    "fmla z23.s, p3/M, z8.s, z12.s\n"
    "ld1w { z12.s }, p2/Z, [x23, x9, LSL #2]\n"
    "fmla z16.s, p3/M, z4.s, z10.s\n"
    "fmla z17.s, p3/M, z3.s, z10.s\n"
    "fmla z18.s, p3/M, z5.s, z11.s\n"
    "ld1w { z9.s }, p0/Z, [x14, x10, LSL #2]\n"
    "fmla z19.s, p3/M, z4.s, z11.s\n"
    "fmla z29.s, p3/M, z8.s, z12.s\n"
    "fmla z30.s, p3/M, z7.s, z12.s\n"
    "fmla z31.s, p3/M, z6.s, z12.s\n"
    "ld1w { z12.s }, p2/Z, [x12, x9, LSL #2]\n"
    "fmla z20.s, p3/M, z1.s, z10.s\n"
    "fmla z21.s, p3/M, z0.s, z10.s\n"
    "ld1w { z10.s }, p2/Z, [x11, x9, LSL #2]\n"
    "ldp x12, x11, [x15, #0x10]\n"
    "fmla z22.s, p3/M, z2.s, z11.s\n"
    "fmla z23.s, p3/M, z1.s, z11.s\n"
    "incw x9\n"
    "ld1w { z11.s }, p0/Z, [x12, x10, LSL #2]\n"
    ".inst 0xc1adc9d0  // fclamp { z16.s-z19.s }, z14.s, z13.s\n"
    "st1w { z16.s }, p1, [x22, x27, LSL #2]\n"
    "ldr x22, [x28, #0x20]\n"
    "fmla z24.s, p3/M, z7.s, z12.s\n"
    "st1w { z17.s }, p1, [x21, x27, LSL #2]\n"
    "ldr x21, [x28, #0x28]\n"
    "fmla z25.s, p3/M, z6.s, z12.s\n"
    "fmla z26.s, p3/M, z8.s, z10.s\n"
    "st1w { z18.s }, p1, [x20, x27, LSL #2]\n"
    "ldr x20, [x28, #0x30]\n"
    "fmla z27.s, p3/M, z7.s, z10.s\n"
    ".inst 0xc1adc9d4  // fclamp { z20.s-z23.s }, z14.s, z13.s\n"
    "st1w { z19.s }, p1, [x19, x27, LSL #2]\n"
    "ldr x19, [x28, #0x38]\n"
    "fmla z28.s, p3/M, z4.s, z12.s\n"
    "fmla z29.s, p3/M, z3.s, z12.s\n"
    "st1w { z20.s }, p1, [x22, x27, LSL #2]\n"
    "ldr x22, [x28, #0x40]\n"
    "fmla z30.s, p3/M, z5.s, z10.s\n"
    "fmla z31.s, p3/M, z4.s, z10.s\n"
    "st1w { z21.s }, p1, [x21, x27, LSL #2]\n"
    "ldr x21, [x28, #0x48]\n"
    ".inst 0xc1adc9d8  // fclamp { z24.s-z27.s }, z14.s, z13.s\n"
    "ld1w { z10.s }, p0/Z, [x13, x10, LSL #2]\n"
    "st1w { z22.s }, p1, [x20, x27, LSL #2]\n"
    "ldr x20, [x28, #0x50]\n"
    "ld1w { z12.s }, p0/Z, [x11, x10, LSL #2]\n"
    "incw x10\n"
    "st1w { z23.s }, p1, [x19, x27, LSL #2]\n"
    "ldr x19, [x28, #0x58]\n"
    ".inst 0xa040c200  // ld1w { z0.s-z3.s }, pn8.b/Z, [x16]\n"
    "addvl x16, x16, #4\n"
    "st1w { z24.s }, p1, [x22, x27, LSL #2]\n"
    "ldr x22, [x28, #0x60]\n"
    "whilelt p2.s, x9, %x[n_channels]\n"
    ".inst 0xa040c204  // ld1w { z4.s-z7.s }, pn8.b/Z, [x16]\n"
    "st1w { z25.s }, p1, [x21, x27, LSL #2]\n"
    "ldr x21, [x28, #0x68]\n"
    "addvl x16, x16, #4\n"
    "cmp x10, %x[n_channels]\n"
    "st1w { z26.s }, p1, [x20, x27, LSL #2]\n"
    "ldr x20, [x28, #0x70]\n"
    ".inst 0xc1adc9dc  // fclamp { z28.s-z31.s }, z14.s, z13.s\n"
    "ld1w { z8.s }, p3/Z, [x16]\n"
    "st1w { z27.s }, p1, [x19, x27, LSL #2]\n"
    "ldr x19, [x28, #0x78]\n"
    "addvl x16, x16, #1\n"
    "st1w { z28.s }, p1, [x22, x27, LSL #2]\n"
    "st1w { z29.s }, p1, [x21, x27, LSL #2]\n"
    "st1w { z30.s }, p1, [x20, x27, LSL #2]\n"
    "st1w { z31.s }, p1, [x19, x27, LSL #2]\n"
    "blt 1b\n"
    "2:"  // Channel tail
    "movprfx z21, z15\n fmla z21.s, p3/M, z4.s, z9.s\n"
    "movprfx z16, z15\n fmla z16.s, p3/M, z8.s, z9.s\n"
    "ldr x26, [x15, #0x20]\n"
    "incw x27\n"
    "movprfx z22, z15\n fmla z22.s, p3/M, z3.s, z9.s\n"
    "movprfx z25, z15\n fmla z25.s, p3/M, z1.s, z9.s\n"
    "ldr x25, [x15, #0x30]\n"
    "mov p1.b, p2.b\n"
    "movprfx z26, z15\n fmla z26.s, p3/M, z0.s, z9.s\n"
    "ldr x24, [x15, #0x28]\n"
    "movprfx z17, z15\n fmla z17.s, p3/M, z7.s, z9.s\n"
    "movprfx z18, z15\n fmla z18.s, p3/M, z6.s, z9.s\n"
    "fmla z21.s, p3/M, z5.s, z12.s\n"
    "ldr x23, [x15, #0x38]\n"
    "movprfx z20, z15\n fmla z20.s, p3/M, z5.s, z9.s\n"
    "movprfx z24, z15\n fmla z24.s, p3/M, z2.s, z9.s\n"
    "ld1w { z9.s }, p2/Z, [x25, x9, LSL #2]\n"
    "ldr x14, [x15, #0x40]\n"
    "fmla z16.s, p3/M, z0.s, z10.s\n"
    "movprfx z19, z15\n fmla z19.s, p3/M, z2.s, z11.s\n"
    "ld1w { z10.s }, p2/Z, [x26, x9, LSL #2]\n"
    "ldr x13, [x15, #0x48]\n"
    "fmla z22.s, p3/M, z4.s, z12.s\n"
    "fmla z25.s, p3/M, z2.s, z12.s\n"
    "ld1w { z11.s }, p2/Z, [x24, x9, LSL #2]\n"
    "ldr x12, [x15, #0x50]\n"
    "fmla z26.s, p3/M, z1.s, z12.s\n"
    "fmla z17.s, p3/M, z8.s, z12.s\n"
    "ldr x26, [x15, #0x60]\n"
    "fmla z18.s, p3/M, z7.s, z12.s\n"
    "movprfx z28, z15\n fmla z28.s, p3/M, z6.s, z10.s\n"
    "ld1w { z10.s }, p2/Z, [x13, x9, LSL #2]\n"
    "ldr x24, [x15, #0x68]\n"
    "fmla z21.s, p3/M, z7.s, z9.s\n"
    "fmla z19.s, p3/M, z6.s, z12.s\n"
    "ldr x11, [x15, #0x58]\n"
    "movprfx z23, z15\n fmla z23.s, p3/M, z3.s, z12.s\n"
    "movprfx z27, z15\n fmla z27.s, p3/M, z0.s, z12.s\n"
    "ld1w { z12.s }, p2/Z, [x23, x9, LSL #2]\n"
    "ldr x25, [x15, #0x70]\n"
    "movprfx z31, z15\n fmla z31.s, p3/M, z8.s, z11.s\n"
    "fmla z22.s, p3/M, z6.s, z9.s\n"
    "ld1w { z11.s }, p2/Z, [x14, x9, LSL #2]\n"
    "ldr x23, [x15, #0x78]\n"
    "fmla z25.s, p3/M, z4.s, z9.s\n"
    "fmla z26.s, p3/M, z3.s, z9.s\n"
    "ldr x14, [x15, #0x80]\n"
    "movprfx z29, z15\n fmla z29.s, p3/M, z1.s, z9.s\n"
    "movprfx z30, z15\n fmla z30.s, p3/M, z0.s, z9.s\n"
    "ldr x13, [x15, #0x88]\n"
    "fmla z20.s, p3/M, z8.s, z9.s\n"
    "fmla z24.s, p3/M, z5.s, z9.s\n"
    "ldr x22, [x28, #0x0]\n"
    "fmla z28.s, p3/M, z2.s, z9.s\n"
    "fmla z16.s, p3/M, z1.s, z12.s\n"
    "ld1w { z9.s }, p2/Z, [x12, x9, LSL #2]\n"
    "ldr x12, [x15, #0x90]\n"
    "fmla z17.s, p3/M, z0.s, z12.s\n"
    "fmla z18.s, p3/M, z2.s, z11.s\n"
    "ld1w { z12.s }, p2/Z, [x11, x9, LSL #2]\n"
    "ldr x11, [x15, #0x98]\n"
    "fmla z21.s, p3/M, z8.s, z10.s\n"
    "fmla z19.s, p3/M, z1.s, z11.s\n"
    "ld1w { z11.s }, p2/Z, [x26, x9, LSL #2]\n"
    "ldr x26, [x15, #0xa0]\n"
    "fmla z22.s, p3/M, z7.s, z10.s\n"
    "fmla z23.s, p3/M, z6.s, z10.s\n"
    "ldr x21, [x28, #0x8]\n"
    "fmla z25.s, p3/M, z5.s, z10.s\n"
    "fmla z26.s, p3/M, z4.s, z10.s\n"
    "ldr x20, [x28, #0x10]\n"
    "fmla z27.s, p3/M, z3.s, z10.s\n"
    "fmla z29.s, p3/M, z2.s, z10.s\n"
    "ldr x19, [x28, #0x18]\n"
    "fmla z30.s, p3/M, z1.s, z10.s\n"
    "fmla z31.s, p3/M, z0.s, z10.s\n"
    "ld1w { z10.s }, p2/Z, [x24, x9, LSL #2]\n"
    "ldr x24, [x15, #0xa8]\n"
    "fmla z16.s, p3/M, z3.s, z9.s\n"
    "fmla z20.s, p3/M, z0.s, z9.s\n"
    "fmla z24.s, p3/M, z6.s, z11.s\n"
    "fmla z28.s, p3/M, z3.s, z11.s\n"
    "ld1w { z11.s }, p2/Z, [x25, x9, LSL #2]\n"
    "ldr x25, [x15, #0xb0]\n"
    "fmla z17.s, p3/M, z4.s, z10.s\n"
    "fmla z18.s, p3/M, z3.s, z10.s\n"
    "fmla z21.s, p3/M, z1.s, z10.s\n"
    "fmla z19.s, p3/M, z5.s, z12.s\n"
    "fmla z23.s, p3/M, z2.s, z12.s\n"
    "fmla z22.s, p3/M, z0.s, z10.s\n"
    "ld1w { z12.s }, p2/Z, [x23, x9, LSL #2]\n"
    "ldr x23, [x15, #0xb8]\n"
    "fmla z27.s, p3/M, z8.s, z11.s\n"
    "fmla z31.s, p3/M, z5.s, z11.s\n"
    "ld1w { z11.s }, p2/Z, [x14, x9, LSL #2]\n"
    "ldr x14, [x15, #0xc0]\n"
    "fmla z16.s, p3/M, z5.s, z10.s\n"
    "fmla z20.s, p3/M, z2.s, z10.s\n"
    "ld1w { z10.s }, p2/Z, [x13, x9, LSL #2]\n"
    "ldr x13, [x15, #0xc8]\n"
    "fmla z17.s, p3/M, z5.s, z12.s\n"
    "fmla z18.s, p3/M, z4.s, z12.s\n"
    "fmla z21.s, p3/M, z2.s, z12.s\n"
    "fmla z19.s, p3/M, z3.s, z12.s\n"
    "fmla z22.s, p3/M, z1.s, z12.s\n"
    "fmla z23.s, p3/M, z0.s, z12.s\n"
    "ld1w { z12.s }, p2/Z, [x11, x9, LSL #2]\n"
    "ldr x11, [x15, #0xd8]\n"
    "fmla z28.s, p3/M, z7.s, z11.s\n"
    "fmla z29.s, p3/M, z6.s, z11.s\n"
    "ld1w { z11.s }, p2/Z, [x12, x9, LSL #2]\n"
    "ldr x12, [x15, #0xd0]\n"
    "fmla z16.s, p3/M, z7.s, z10.s\n"
    "fmla z17.s, p3/M, z6.s, z10.s\n"
    "fmla z20.s, p3/M, z4.s, z10.s\n"
    "fmla z21.s, p3/M, z3.s, z10.s\n"
    "fmla z24.s, p3/M, z1.s, z10.s\n"
    "fmla z25.s, p3/M, z0.s, z10.s\n"
    "ld1w { z10.s }, p2/Z, [x26, x9, LSL #2]\n"
    "ldr x26, [x15, #0xe0]\n"
    "fmla z18.s, p3/M, z8.s, z12.s\n"
    "fmla z30.s, p3/M, z8.s, z11.s\n"
    "fmla z31.s, p3/M, z7.s, z11.s\n"
    "ld1w { z11.s }, p2/Z, [x24, x9, LSL #2]\n"
    "fmla z27.s, p3/M, z1.s, z12.s\n"
    "ldr x24, [x15, #0xe8]\n"
    "fmla z19.s, p3/M, z7.s, z12.s\n"
    "fmla z22.s, p3/M, z5.s, z12.s\n"
    "fmla z23.s, p3/M, z4.s, z12.s\n"
    "fmla z26.s, p3/M, z2.s, z12.s\n"
    "ld1w { z12.s }, p2/Z, [x25, x9, LSL #2]\n"
    "ldr x25, [x15, #0xf0]\n"
    "fmla z16.s, p3/M, z2.s, z10.s\n"
    "fmla z17.s, p3/M, z1.s, z10.s\n"
    "fmla z18.s, p3/M, z0.s, z10.s\n"
    "fmla z20.s, p3/M, z7.s, z11.s\n"
    "ld1w { z10.s }, p2/Z, [x23, x9, LSL #2]\n"
    "ldr x23, [x15, #0xf8]\n"
    "fmla z21.s, p3/M, z6.s, z11.s\n"
    "fmla z24.s, p3/M, z4.s, z11.s\n"
    "fmla z25.s, p3/M, z3.s, z11.s\n"
    "fmla z28.s, p3/M, z1.s, z11.s\n"
    "fmla z29.s, p3/M, z0.s, z11.s\n"
    "ld1w { z11.s }, p2/Z, [x14, x9, LSL #2]\n"
    "fmla z27.s, p3/M, z4.s, z11.s\n"
    "ldr x14, [x15, #0x100]\n"
    "fmla z30.s, p3/M, z2.s, z11.s\n"
    "fmla z17.s, p3/M, z2.s, z12.s\n"
    "fmla z18.s, p3/M, z1.s, z12.s\n"
    "fmla z19.s, p3/M, z0.s, z12.s\n"
    "ld1w { z12.s }, p2/Z, [x13, x9, LSL #2]\n"
    "ldr x13, [x15, #0x108]\n"
    "fmla z16.s, p3/M, z6.s, z10.s\n"
    "fmla z20.s, p3/M, z3.s, z10.s\n"
    "fmla z24.s, p3/M, z0.s, z10.s\n"
    "fmla z22.s, p3/M, z8.s, z11.s\n"
    "ld1w { z10.s }, p2/Z, [x12, x9, LSL #2]\n"
    "ldr x12, [x15, #0x110]\n"
    "fmla z23.s, p3/M, z7.s, z11.s\n"
    "fmla z26.s, p3/M, z5.s, z11.s\n"
    "fmla z31.s, p3/M, z1.s, z11.s\n"
    "ld1w { z11.s }, p2/Z, [x11, x9, LSL #2]\n"
    "fmla z27.s, p3/M, z2.s, z12.s\n"
    "ldr x11, [x15, #0x118]\n"
    "fmla z28.s, p3/M, z0.s, z10.s\n"
    "fmla z29.s, p3/M, z4.s, z11.s\n"
    "fmla z30.s, p3/M, z3.s, z11.s\n"
    "fmla z19.s, p3/M, z8.s, z12.s\n"
    "fmla z23.s, p3/M, z5.s, z12.s\n"
    "fmla z20.s, p3/M, z6.s, z10.s\n"
    "ld1w { z12.s }, p2/Z, [x26, x9, LSL #2]\n"
    "fmla z24.s, p3/M, z3.s, z10.s\n"
    "ld1w { z10.s }, p2/Z, [x24, x9, LSL #2]\n"
    "fmla z25.s, p3/M, z7.s, z11.s\n"
    "fmla z26.s, p3/M, z6.s, z11.s\n"
    "fmla z28.s, p3/M, z5.s, z11.s\n"
    "fmla z27.s, p3/M, z5.s, z12.s\n"
    "fmla z31.s, p3/M, z2.s, z12.s\n"
    "fmla z29.s, p3/M, z7.s, z10.s\n"
    "fmla z30.s, p3/M, z6.s, z10.s\n"
    "fmla z24.s, p3/M, z8.s, z11.s\n"
    "ld1w { z11.s }, p2/Z, [x25, x9, LSL #2]\n"
    "fmla z28.s, p3/M, z8.s, z10.s\n"
    "fmla z25.s, p3/M, z8.s, z11.s\n"
    "fmla z26.s, p3/M, z7.s, z11.s\n"
    "ld1w { z10.s }, p2/Z, [x14, x9, LSL #2]\n"
    "fmla z27.s, p3/M, z6.s, z11.s\n"
    "fmla z29.s, p3/M, z5.s, z11.s\n"
    "fmla z30.s, p3/M, z4.s, z11.s\n"
    "fmla z31.s, p3/M, z3.s, z11.s\n"
    "ld1w { z11.s }, p2/Z, [x13, x9, LSL #2]\n"
    "fmla z23.s, p3/M, z8.s, z12.s\n"
    "ld1w { z12.s }, p2/Z, [x23, x9, LSL #2]\n"
    "fmla z16.s, p3/M, z4.s, z10.s\n"
    "fmla z17.s, p3/M, z3.s, z10.s\n"
    "fmla z18.s, p3/M, z5.s, z11.s\n"
    "fmla z19.s, p3/M, z4.s, z11.s\n"
    "fmla z29.s, p3/M, z8.s, z12.s\n"
    "fmla z30.s, p3/M, z7.s, z12.s\n"
    "fmla z31.s, p3/M, z6.s, z12.s\n"
    "ld1w { z12.s }, p2/Z, [x12, x9, LSL #2]\n"
    "fmla z20.s, p3/M, z1.s, z10.s\n"
    "fmla z21.s, p3/M, z0.s, z10.s\n"
    "ld1w { z10.s }, p2/Z, [x11, x9, LSL #2]\n"
    "fmla z22.s, p3/M, z2.s, z11.s\n"
    "fmla z23.s, p3/M, z1.s, z11.s\n"
    ".inst 0xc1adc9d0  // fclamp { z16.s-z19.s }, z14.s, z13.s\n"
    "st1w { z16.s }, p1, [x22, x27, LSL #2]\n"
    "ldr x22, [x28, #0x20]\n"
    "fmla z24.s, p3/M, z7.s, z12.s\n"
    "st1w { z17.s }, p1, [x21, x27, LSL #2]\n"
    "ldr x21, [x28, #0x28]\n"
    "fmla z25.s, p3/M, z6.s, z12.s\n"
    "fmla z26.s, p3/M, z8.s, z10.s\n"
    "st1w { z18.s }, p1, [x20, x27, LSL #2]\n"
    "ldr x20, [x28, #0x30]\n"
    "fmla z27.s, p3/M, z7.s, z10.s\n"
    ".inst 0xc1adc9d4  // fclamp { z20.s-z23.s }, z14.s, z13.s\n"
    "st1w { z19.s }, p1, [x19, x27, LSL #2]\n"
    "ldr x19, [x28, #0x38]\n"
    "fmla z28.s, p3/M, z4.s, z12.s\n"
    "fmla z29.s, p3/M, z3.s, z12.s\n"
    "st1w { z20.s }, p1, [x22, x27, LSL #2]\n"
    "ldr x22, [x28, #0x40]\n"
    "fmla z30.s, p3/M, z5.s, z10.s\n"
    "fmla z31.s, p3/M, z4.s, z10.s\n"
    "st1w { z21.s }, p1, [x21, x27, LSL #2]\n"
    "ldr x21, [x28, #0x48]\n"
    ".inst 0xc1adc9d8  // fclamp { z24.s-z27.s }, z14.s, z13.s\n"
    ".inst 0xc1adc9dc  // fclamp { z28.s-z31.s }, z14.s, z13.s\n"
    "st1w { z22.s }, p1, [x20, x27, LSL #2]\n"
    "ldr x20, [x28, #0x50]\n"
    "st1w { z23.s }, p1, [x19, x27, LSL #2]\n"
    "ldr x19, [x28, #0x58]\n"
    "st1w { z24.s }, p1, [x22, x27, LSL #2]\n"
    "ldr x22, [x28, #0x60]\n"
    "st1w { z25.s }, p1, [x21, x27, LSL #2]\n"
    "ldr x21, [x28, #0x68]\n"
    "st1w { z26.s }, p1, [x20, x27, LSL #2]\n"
    "ldr x20, [x28, #0x70]\n"
    "st1w { z27.s }, p1, [x19, x27, LSL #2]\n"
    "ldr x19, [x28, #0x78]\n"
    "st1w { z28.s }, p1, [x22, x27, LSL #2]\n"
    "st1w { z29.s }, p1, [x21, x27, LSL #2]\n"
    "st1w { z30.s }, p1, [x20, x27, LSL #2]\n"
    "st1w { z31.s }, p1, [x19, x27, LSL #2]\n"
    ".inst 0xd503467f  // SMSTOP\n"
    :
    : [n_channels] "r" ((unsigned long) n_channels), [offsetof_Args_inptrs] "I" (offsetof(Args, inptrs)), [offsetof_args_max] "I" (offsetof(Args, max)), [offsetof_args_min] "I" (offsetof(Args, min)), [offsetof_args_outptrs] "I" (offsetof(Args, outptrs)), [offsetof_args_params] "I" (offsetof(Args, params)), [params_struct] "r" (&params_struct)
    : "cc", "memory", "p0", "p1", "p2", "p3", "p4", "p5", "p6", "p7", "p8", "p9", "p10", "p11", "p12", "p13", "p14", "p15", "x9", "x10", "x11", "x12", "x13", "x14", "x15", "x16", "x19", "x20", "x21", "x22", "x23", "x24", "x25", "x26", "x27", "x28", "z0", "z1", "z2", "z3", "z4", "z5", "z6", "z7", "z8", "z9", "z10", "z11", "z12", "z13", "z14", "z15", "z16", "z17", "z18", "z19", "z20", "z21", "z22", "z23", "z24", "z25", "z26", "z27", "z28", "z29", "z30", "z31"
  );
}

}  // namespace depthwise
}  // namespace arm_conv

#endif  // defined(ARM_COMPUTE_ENABLE_SME2)
