/*
 * Copyright (c) 2021-2022 Arm Limited.
 *
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#if defined(ARM_COMPUTE_ENABLE_SVE)

#include "arm_gemm.hpp"
#include <cstdint>

namespace arm_conv {
namespace depthwise {

void sve_u8q_nhwc_3x3_s1_output2x2_dot_depthfirst_impl(
  const unsigned int n_channels,
  const uint8_t *const *const inptrs,
  const uint8_t *params,
  const int32_t *,  // Bias, should be wrapped into the parameters
  const arm_gemm::Requantize32& qp,
  const int32_t *, const int32_t *,  // Requant parameters, also wrapped
  uint8_t *const *const outptrs
)
{
  __asm__ __volatile__(
    "ldp x11, x10, [%x[inptrs], #0x0]\n"
    "ptrue p2.b\n"
    "ldp x9, x28, [%x[inptrs], #0x10]\n"
    "addvl SP, SP, #-8\n"
    "ldp x27, x26, [%x[inptrs], #0x20]\n"
    "mov x19, #0x1\n"
    "ldp x25, x24, [%x[inptrs], #0x30]\n"
    "orr x19, x19, #0x100\n"
    "ldp x23, x22, [%x[outptrs], #0x0]\n"
    "orr x19, x19, #0x10000\n"
    "dup z12.s, w19\n"
    "ldp x21, x20, [%x[outptrs], #0x10]\n"
    "mov x19, #0x0\n"
    "ld1rw { z11.s }, p2/Z, [%x[qp], %[offsetof_Requantize32_minval]]\n"
    "whilelt p1.b, x19, %x[n_channels]\n"
    "ld1rw { z10.s }, p2/Z, [%x[qp], %[offsetof_Requantize32_maxval]]\n"
    "ld1rw { z9.s }, p2/Z, [%x[qp], %[offsetof_Requantize32_b_offset]]\n"
    "ld1rw { z8.s }, p2/Z, [%x[qp], %[offsetof_Requantize32_c_offset]]\n"
    "1:"  // Loop
    "mov z7.s, #0x0\n"
    "ld1b { z19.b }, p1/Z, [x11, x19]\n"
    "whilelt p0.s, x19, %x[n_channels]\n"
    "mov z6.s, #0x0\n"
    "ld1b { z18.b }, p1/Z, [x10, x19]\n"
    "ldp x11, x10, [%x[inptrs], #0x40]\n"
    "ld1b { z16.b }, p1/Z, [x9, x19]\n"
    "zip1 z21.b, z19.b, z16.b\n"
    "ld1b { z17.b }, p1/Z, [x28, x19]\n"
    "zip2 z19.b, z19.b, z16.b\n"
    "ldp x9, x28, [%x[inptrs], #0x50]\n"
    "ld1b { z23.b }, p1/Z, [x27, x19]\n"
    "zip1 z16.b, z18.b, z17.b\n"
    "ld1b { z20.b }, p1/Z, [x26, x19]\n"
    "zip2 z18.b, z18.b, z17.b\n"
    "ldp x27, x26, [%x[inptrs], #0x60]\n"
    "zip1 z5.b, z21.b, z16.b\n"
    "ld1b { z17.b }, p1/Z, [x25, x19]\n"
    "zip2 z4.b, z21.b, z16.b\n"
    "ld1b { z16.b }, p1/Z, [x24, x19]\n"
    "zip1 z29.b, z19.b, z18.b\n"
    "ldp x25, x24, [%x[inptrs], #0x70]\n"
    "zip2 z28.b, z19.b, z18.b\n"
    "ld1b { z22.b }, p1/Z, [x11, x19]\n"
    "zip1 z19.b, z23.b, z17.b\n"
    "ld1b { z21.b }, p1/Z, [x10, x19]\n"
    "zip2 z27.b, z23.b, z17.b\n"
    "ldp x11, x10, [%x[inptrs], #0x0]\n"
    "zip1 z18.b, z20.b, z16.b\n"
    "ld1b { z17.b }, p1/Z, [x9, x19]\n"
    "zip2 z20.b, z20.b, z16.b\n"
    "ld1b { z16.b }, p1/Z, [x28, x19]\n"
    "zip1 z3.b, z19.b, z18.b\n"
    "ldp x9, x28, [%x[inptrs], #0x10]\n"
    "zip2 z2.b, z19.b, z18.b\n"
    "ld1b { z19.b }, p1/Z, [x27, x19]\n"
    "zip1 z26.b, z22.b, z17.b\n"
    "ld1b { z25.b }, p1/Z, [x26, x19]\n"
    "zip2 z24.b, z22.b, z17.b\n"
    "ldp x27, x26, [%x[inptrs], #0x20]\n"
    "zip1 z23.b, z21.b, z16.b\n"
    "ld1b { z18.b }, p1/Z, [x25, x19]\n"
    "zip2 z22.b, z21.b, z16.b\n"
    "ld1b { z21.b }, p1/Z, [x24, x19]\n"
    "zip1 z17.b, z27.b, z20.b\n"
    "ldp x25, x24, [%x[inptrs], #0x30]\n"
    "zip2 z16.b, z27.b, z20.b\n"
    "st1b { z29.b }, p2, [SP]\n"
    "zip1 z20.b, z19.b, z18.b\n"
    "st1b { z28.b }, p2, [SP, #1, MUL VL]\n"
    "zip2 z19.b, z19.b, z18.b\n"
    "st1b { z17.b }, p2, [SP, #2, MUL VL]\n"
    "zip1 z18.b, z25.b, z21.b\n"
    "st1b { z16.b }, p2, [SP, #3, MUL VL]\n"
    "zip2 z17.b, z25.b, z21.b\n"
    "ld1w { z1.s }, p2/Z, [%x[params]]\n"
    "zip1 z0.b, z26.b, z23.b\n"
    "ld1b { z31.b }, p2/Z, [%x[params], #1, MUL VL]\n"
    "zip2 z30.b, z26.b, z23.b\n"
    "ld1b { z29.b }, p2/Z, [%x[params], #2, MUL VL]\n"
    "zip1 z16.b, z24.b, z22.b\n"
    "st1b { z16.b }, p2, [SP, #4, MUL VL]\n"
    "zip2 z16.b, z24.b, z22.b\n"
    "st1b { z16.b }, p2, [SP, #5, MUL VL]\n"
    "zip1 z28.b, z20.b, z18.b\n"
    "ld1b { z27.b }, p2/Z, [%x[params], #3, MUL VL]\n"
    "zip2 z26.b, z20.b, z18.b\n"
    "ld1w { z25.s }, p2/Z, [%x[params], #4, MUL VL]\n"
    "zip1 z16.b, z19.b, z17.b\n"
    "st1b { z16.b }, p2, [SP, #6, MUL VL]\n"
    "zip2 z16.b, z19.b, z17.b\n"
    "st1b { z16.b }, p2, [SP, #7, MUL VL]\n"
    "mov z24.d, z1.d\n"
    "ld1w { z23.s }, p2/Z, [%x[params], #5, MUL VL]\n"
    "mov z22.d, z1.d\n"
    "mov z21.d, z1.d\n"
    "udot z1.s, z31.b, z5.b\n"
    "udot z22.s, z31.b, z3.b\n"
    "udot z7.s, z12.b, z3.b\n"
    "udot z1.s, z29.b, z3.b\n"
    "ext z3.b, z3.b, z3.b, #0x1\n"
    "udot z22.s, z29.b, z0.b\n"
    "udot z7.s, z12.b, z0.b\n"
    "udot z1.s, z27.b, z0.b\n"
    "ext z0.b, z0.b, z0.b, #0x1\n"
    "udot z22.s, z27.b, z28.b\n"
    "mov z20.d, z7.d\n"
    "udot z7.s, z12.b, z5.b\n"
    "udot z20.s, z12.b, z28.b\n"
    "ext z5.b, z5.b, z5.b, #0x1\n"
    "ext z28.b, z28.b, z28.b, #0x1\n"
    "udot z21.s, z31.b, z3.b\n"
    "udot z6.s, z12.b, z3.b\n"
    "udot z24.s, z31.b, z5.b\n"
    "ld1b { z31.b }, p2/Z, [%x[params], #7, MUL VL]\n"
    "mls z1.s, p2/M, z7.s, z9.s\n"
    "udot z21.s, z29.b, z0.b\n"
    "udot z6.s, z12.b, z0.b\n"
    "udot z24.s, z29.b, z3.b\n"
    "ld1b { z3.b }, p2/Z, [SP, #2, MUL VL]\n"
    ".inst 0x04b97421  // sqrdmulh z1.s, z1.s, z25.s\n"
    "udot z21.s, z27.b, z28.b\n"
    "mov z19.d, z6.d\n"
    "udot z24.s, z27.b, z0.b\n"
    "ld1b { z0.b }, p2/Z, [SP, #4, MUL VL]\n"
    "udot z6.s, z12.b, z5.b\n"
    "ld1b { z5.b }, p2/Z, [SP]\n"
    "udot z19.s, z12.b, z28.b\n"
    "ld1b { z28.b }, p2/Z, [SP, #6, MUL VL]\n"
    "and z16.d, z1.d, z23.d\n"
    "asr z16.s, z16.s, #0x1f\n"
    "mov z7.s, #0x0\n"
    "mls z24.s, p2/M, z6.s, z9.s\n"
    "udot z7.s, z12.b, z2.b\n"
    "mov z6.s, #0x0\n"
    "mls z22.s, p2/M, z20.s, z9.s\n"
    ".inst 0x04b97718  // sqrdmulh z24.s, z24.s, z25.s\n"
    "sqadd z1.s, z1.s, z16.s\n"
    "udot z7.s, z12.b, z30.b\n"
    ".inst 0x04b976d6  // sqrdmulh z22.s, z22.s, z25.s\n"
    "and z18.d, z24.d, z23.d\n"
    "asr z18.s, z18.s, #0x1f\n"
    "and z17.d, z22.d, z23.d\n"
    "mov z20.d, z7.d\n"
    "asr z17.s, z17.s, #0x1f\n"
    "udot z7.s, z12.b, z4.b\n"
    "udot z20.s, z12.b, z26.b\n"
    "mls z21.s, p2/M, z19.s, z9.s\n"
    "sqadd z24.s, z24.s, z18.s\n"
    ".inst 0x44828ae1  // srshl z1.s, p2/M, z1.s, z23.s\n"
    "sqadd z22.s, z22.s, z17.s\n"
    ".inst 0x04b976b5  // sqrdmulh z21.s, z21.s, z25.s\n"
    ".inst 0x44828af8  // srshl z24.s, p2/M, z24.s, z23.s\n"
    "add z1.s, z1.s, z8.s\n"
    "and z16.d, z21.d, z23.d\n"
    "asr z16.s, z16.s, #0x1f\n"
    "add z24.s, z24.s, z8.s\n"
    "smax z1.s, p2/M, z1.s, z11.s\n"
    ".inst 0x44828af6  // srshl z22.s, p2/M, z22.s, z23.s\n"
    "smax z24.s, p2/M, z24.s, z11.s\n"
    "smin z1.s, p2/M, z1.s, z10.s\n"
    "st1b { z1.s }, p0, [x23, x19]\n"
    "add z22.s, z22.s, z8.s\n"
    "sqadd z21.s, z21.s, z16.s\n"
    "ld1w { z1.s }, p2/Z, [%x[params], #6, MUL VL]\n"
    "addvl %x[params], %x[params], #16\n"
    "smin z24.s, p2/M, z24.s, z10.s\n"
    "ld1b { z29.b }, p2/Z, [%x[params], #-8, MUL VL]\n"
    "ld1b { z27.b }, p2/Z, [%x[params], #-7, MUL VL]\n"
    "smax z22.s, p2/M, z22.s, z11.s\n"
    "ld1w { z25.s }, p2/Z, [%x[params], #-6, MUL VL]\n"
    ".inst 0x44828af5  // srshl z21.s, p2/M, z21.s, z23.s\n"
    "ld1w { z23.s }, p2/Z, [%x[params], #-5, MUL VL]\n"
    "smin z22.s, p2/M, z22.s, z10.s\n"
    "st1b { z24.s }, p0, [x22, x19]\n"
    "mov z24.d, z1.d\n"
    "st1b { z22.s }, p0, [x21, x19]\n"
    "add z21.s, z21.s, z8.s\n"
    "mov z22.d, z1.d\n"
    "udot z22.s, z31.b, z2.b\n"
    "smax z21.s, p2/M, z21.s, z11.s\n"
    "udot z22.s, z29.b, z30.b\n"
    "smin z21.s, p2/M, z21.s, z10.s\n"
    "st1b { z21.s }, p0, [x20, x19]\n"
    "mov z21.d, z1.d\n"
    "incw x19\n"
    "udot z1.s, z31.b, z4.b\n"
    "whilelt p0.s, x19, %x[n_channels]\n"
    "udot z22.s, z27.b, z26.b\n"
    "ext z4.b, z4.b, z4.b, #0x1\n"
    "ext z26.b, z26.b, z26.b, #0x1\n"
    "udot z1.s, z29.b, z2.b\n"
    "ext z2.b, z2.b, z2.b, #0x1\n"
    "udot z24.s, z31.b, z4.b\n"
    "mls z22.s, p2/M, z20.s, z9.s\n"
    "udot z1.s, z27.b, z30.b\n"
    "ext z30.b, z30.b, z30.b, #0x1\n"
    "udot z21.s, z31.b, z2.b\n"
    "ld1b { z31.b }, p2/Z, [%x[params], #-3, MUL VL]\n"
    "udot z24.s, z29.b, z2.b\n"
    "udot z6.s, z12.b, z2.b\n"
    "ld1b { z2.b }, p2/Z, [SP, #3, MUL VL]\n"
    ".inst 0x04b976d6  // sqrdmulh z22.s, z22.s, z25.s\n"
    "udot z21.s, z29.b, z30.b\n"
    "ld1b { z29.b }, p2/Z, [%x[params], #-2, MUL VL]\n"
    "udot z24.s, z27.b, z30.b\n"
    "udot z6.s, z12.b, z30.b\n"
    "ld1b { z30.b }, p2/Z, [SP, #5, MUL VL]\n"
    "and z17.d, z22.d, z23.d\n"
    "asr z17.s, z17.s, #0x1f\n"
    "udot z21.s, z27.b, z26.b\n"
    "ld1b { z27.b }, p2/Z, [%x[params], #-1, MUL VL]\n"
    "mov z19.d, z6.d\n"
    "udot z6.s, z12.b, z4.b\n"
    "ld1b { z4.b }, p2/Z, [SP, #1, MUL VL]\n"
    "udot z19.s, z12.b, z26.b\n"
    "ld1b { z26.b }, p2/Z, [SP, #7, MUL VL]\n"
    "mls z1.s, p2/M, z7.s, z9.s\n"
    "mov z7.s, #0x0\n"
    "sqadd z22.s, z22.s, z17.s\n"
    "udot z7.s, z12.b, z3.b\n"
    ".inst 0x04b97421  // sqrdmulh z1.s, z1.s, z25.s\n"
    "mls z24.s, p2/M, z6.s, z9.s\n"
    "mov z6.s, #0x0\n"
    "udot z7.s, z12.b, z0.b\n"
    "and z16.d, z1.d, z23.d\n"
    "asr z16.s, z16.s, #0x1f\n"
    ".inst 0x04b97718  // sqrdmulh z24.s, z24.s, z25.s\n"
    "mov z20.d, z7.d\n"
    "udot z7.s, z12.b, z5.b\n"
    "udot z20.s, z12.b, z28.b\n"
    "mls z21.s, p2/M, z19.s, z9.s\n"
    "and z18.d, z24.d, z23.d\n"
    "asr z18.s, z18.s, #0x1f\n"
    "sqadd z1.s, z1.s, z16.s\n"
    ".inst 0x04b976b5  // sqrdmulh z21.s, z21.s, z25.s\n"
    "ld1w { z25.s }, p2/Z, [%x[params]]\n"
    ".inst 0x44828af6  // srshl z22.s, p2/M, z22.s, z23.s\n"
    "and z16.d, z21.d, z23.d\n"
    "asr z16.s, z16.s, #0x1f\n"
    "sqadd z24.s, z24.s, z18.s\n"
    "add z22.s, z22.s, z8.s\n"
    ".inst 0x44828ae1  // srshl z1.s, p2/M, z1.s, z23.s\n"
    "smax z22.s, p2/M, z22.s, z11.s\n"
    ".inst 0x44828af8  // srshl z24.s, p2/M, z24.s, z23.s\n"
    "add z1.s, z1.s, z8.s\n"
    "sqadd z21.s, z21.s, z16.s\n"
    "smin z22.s, p2/M, z22.s, z10.s\n"
    "st1b { z22.s }, p0, [x21, x19]\n"
    "add z24.s, z24.s, z8.s\n"
    "smax z1.s, p2/M, z1.s, z11.s\n"
    ".inst 0x44828af5  // srshl z21.s, p2/M, z21.s, z23.s\n"
    "ld1w { z23.s }, p2/Z, [%x[params], #1, MUL VL]\n"
    "smax z24.s, p2/M, z24.s, z11.s\n"
    "smin z1.s, p2/M, z1.s, z10.s\n"
    "st1b { z1.s }, p0, [x23, x19]\n"
    "add z21.s, z21.s, z8.s\n"
    "smin z24.s, p2/M, z24.s, z10.s\n"
    "ld1w { z1.s }, p2/Z, [%x[params], #-4, MUL VL]\n"
    "smax z21.s, p2/M, z21.s, z11.s\n"
    "st1b { z24.s }, p0, [x22, x19]\n"
    "mov z24.d, z1.d\n"
    "mov z22.d, z1.d\n"
    "udot z22.s, z31.b, z3.b\n"
    "smin z21.s, p2/M, z21.s, z10.s\n"
    "st1b { z21.s }, p0, [x20, x19]\n"
    "mov z21.d, z1.d\n"
    "incw x19\n"
    "udot z1.s, z31.b, z5.b\n"
    "whilelt p0.s, x19, %x[n_channels]\n"
    "udot z22.s, z29.b, z0.b\n"
    "ext z5.b, z5.b, z5.b, #0x1\n"
    "udot z1.s, z29.b, z3.b\n"
    "udot z22.s, z27.b, z28.b\n"
    "ext z3.b, z3.b, z3.b, #0x1\n"
    "ext z28.b, z28.b, z28.b, #0x1\n"
    "udot z24.s, z31.b, z5.b\n"
    "udot z1.s, z27.b, z0.b\n"
    "ext z0.b, z0.b, z0.b, #0x1\n"
    "udot z21.s, z31.b, z3.b\n"
    "ld1b { z31.b }, p2/Z, [%x[params], #3, MUL VL]\n"
    "udot z24.s, z29.b, z3.b\n"
    "udot z6.s, z12.b, z3.b\n"
    "mls z1.s, p2/M, z7.s, z9.s\n"
    "udot z21.s, z29.b, z0.b\n"
    "ld1b { z29.b }, p2/Z, [%x[params], #4, MUL VL]\n"
    "udot z24.s, z27.b, z0.b\n"
    "udot z6.s, z12.b, z0.b\n"
    ".inst 0x04b97421  // sqrdmulh z1.s, z1.s, z25.s\n"
    "udot z21.s, z27.b, z28.b\n"
    "ld1b { z27.b }, p2/Z, [%x[params], #5, MUL VL]\n"
    "mov z7.s, #0x0\n"
    "mov z19.d, z6.d\n"
    "udot z6.s, z12.b, z5.b\n"
    "udot z19.s, z12.b, z28.b\n"
    "and z16.d, z1.d, z23.d\n"
    "asr z16.s, z16.s, #0x1f\n"
    "udot z7.s, z12.b, z2.b\n"
    "mls z24.s, p2/M, z6.s, z9.s\n"
    "mov z6.s, #0x0\n"
    "mls z22.s, p2/M, z20.s, z9.s\n"
    "mls z21.s, p2/M, z19.s, z9.s\n"
    ".inst 0x04b97718  // sqrdmulh z24.s, z24.s, z25.s\n"
    "sqadd z1.s, z1.s, z16.s\n"
    ".inst 0x04b976d6  // sqrdmulh z22.s, z22.s, z25.s\n"
    ".inst 0x04b976b5  // sqrdmulh z21.s, z21.s, z25.s\n"
    "ld1w { z25.s }, p2/Z, [%x[params], #6, MUL VL]\n"
    "and z18.d, z24.d, z23.d\n"
    "asr z18.s, z18.s, #0x1f\n"
    "and z17.d, z22.d, z23.d\n"
    "and z16.d, z21.d, z23.d\n"
    "asr z17.s, z17.s, #0x1f\n"
    "udot z7.s, z12.b, z30.b\n"
    ".inst 0x44828ae1  // srshl z1.s, p2/M, z1.s, z23.s\n"
    "asr z16.s, z16.s, #0x1f\n"
    "sqadd z24.s, z24.s, z18.s\n"
    "add z1.s, z1.s, z8.s\n"
    "mov z20.d, z7.d\n"
    "sqadd z22.s, z22.s, z17.s\n"
    "sqadd z21.s, z21.s, z16.s\n"
    "udot z7.s, z12.b, z4.b\n"
    "udot z20.s, z12.b, z26.b\n"
    "smax z1.s, p2/M, z1.s, z11.s\n"
    ".inst 0x44828af8  // srshl z24.s, p2/M, z24.s, z23.s\n"
    ".inst 0x44828af6  // srshl z22.s, p2/M, z22.s, z23.s\n"
    ".inst 0x44828af5  // srshl z21.s, p2/M, z21.s, z23.s\n"
    "ld1w { z23.s }, p2/Z, [%x[params], #7, MUL VL]\n"
    "smin z1.s, p2/M, z1.s, z10.s\n"
    "st1b { z1.s }, p0, [x23, x19]\n"
    "add z24.s, z24.s, z8.s\n"
    "add z22.s, z22.s, z8.s\n"
    "ld1w { z1.s }, p2/Z, [%x[params], #2, MUL VL]\n"
    "addvl %x[params], %x[params], #8\n"
    "add z21.s, z21.s, z8.s\n"
    "smax z24.s, p2/M, z24.s, z11.s\n"
    "smax z22.s, p2/M, z22.s, z11.s\n"
    "smax z21.s, p2/M, z21.s, z11.s\n"
    "smin z24.s, p2/M, z24.s, z10.s\n"
    "st1b { z24.s }, p0, [x22, x19]\n"
    "mov z24.d, z1.d\n"
    "smin z22.s, p2/M, z22.s, z10.s\n"
    "st1b { z22.s }, p0, [x21, x19]\n"
    "mov z22.d, z1.d\n"
    "smin z21.s, p2/M, z21.s, z10.s\n"
    "st1b { z21.s }, p0, [x20, x19]\n"
    "mov z21.d, z1.d\n"
    "incw x19\n"
    "udot z1.s, z31.b, z4.b\n"
    "whilelt p0.s, x19, %x[n_channels]\n"
    "udot z22.s, z31.b, z2.b\n"
    "ext z4.b, z4.b, z4.b, #0x1\n"
    "udot z1.s, z29.b, z2.b\n"
    "udot z22.s, z29.b, z30.b\n"
    "ext z2.b, z2.b, z2.b, #0x1\n"
    "udot z24.s, z31.b, z4.b\n"
    "udot z1.s, z27.b, z30.b\n"
    "udot z22.s, z27.b, z26.b\n"
    "ext z30.b, z30.b, z30.b, #0x1\n"
    "ext z26.b, z26.b, z26.b, #0x1\n"
    "udot z21.s, z31.b, z2.b\n"
    "udot z24.s, z29.b, z2.b\n"
    "udot z6.s, z12.b, z2.b\n"
    "mls z1.s, p2/M, z7.s, z9.s\n"
    "udot z21.s, z29.b, z30.b\n"
    "udot z24.s, z27.b, z30.b\n"
    "udot z6.s, z12.b, z30.b\n"
    ".inst 0x04b97421  // sqrdmulh z1.s, z1.s, z25.s\n"
    "udot z21.s, z27.b, z26.b\n"
    "mls z22.s, p2/M, z20.s, z9.s\n"
    "mov z19.d, z6.d\n"
    "udot z6.s, z12.b, z4.b\n"
    "udot z19.s, z12.b, z26.b\n"
    "and z16.d, z1.d, z23.d\n"
    "asr z16.s, z16.s, #0x1f\n"
    ".inst 0x04b976d6  // sqrdmulh z22.s, z22.s, z25.s\n"
    "mls z24.s, p2/M, z6.s, z9.s\n"
    "mls z21.s, p2/M, z19.s, z9.s\n"
    ".inst 0x04b97718  // sqrdmulh z24.s, z24.s, z25.s\n"
    "and z17.d, z22.d, z23.d\n"
    "asr z17.s, z17.s, #0x1f\n"
    "sqadd z1.s, z1.s, z16.s\n"
    ".inst 0x04b976b5  // sqrdmulh z21.s, z21.s, z25.s\n"
    "and z18.d, z24.d, z23.d\n"
    "asr z18.s, z18.s, #0x1f\n"
    "and z16.d, z21.d, z23.d\n"
    ".inst 0x44828ae1  // srshl z1.s, p2/M, z1.s, z23.s\n"
    "asr z16.s, z16.s, #0x1f\n"
    "sqadd z22.s, z22.s, z17.s\n"
    "add z1.s, z1.s, z8.s\n"
    "sqadd z24.s, z24.s, z18.s\n"
    "smax z1.s, p2/M, z1.s, z11.s\n"
    ".inst 0x44828af6  // srshl z22.s, p2/M, z22.s, z23.s\n"
    "sqadd z21.s, z21.s, z16.s\n"
    ".inst 0x44828af8  // srshl z24.s, p2/M, z24.s, z23.s\n"
    "add z22.s, z22.s, z8.s\n"
    "smin z1.s, p2/M, z1.s, z10.s\n"
    "st1b { z1.s }, p0, [x23, x19]\n"
    "add z24.s, z24.s, z8.s\n"
    "smax z22.s, p2/M, z22.s, z11.s\n"
    ".inst 0x44828af5  // srshl z21.s, p2/M, z21.s, z23.s\n"
    "smax z24.s, p2/M, z24.s, z11.s\n"
    "smin z22.s, p2/M, z22.s, z10.s\n"
    "st1b { z22.s }, p0, [x21, x19]\n"
    "add z21.s, z21.s, z8.s\n"
    "smin z24.s, p2/M, z24.s, z10.s\n"
    "st1b { z24.s }, p0, [x22, x19]\n"
    "smax z21.s, p2/M, z21.s, z11.s\n"
    "smin z21.s, p2/M, z21.s, z10.s\n"
    "st1b { z21.s }, p0, [x20, x19]\n"
    "incw x19\n"
    "whilelt p1.b, x19, %x[n_channels]\n"
    "b.any 1b\n"
    "addvl SP, SP, #8\n"
    : [params] "+&r" (params)
    : [inptrs] "r" (inptrs), [n_channels] "r" ((long unsigned int) n_channels), [offsetof_Requantize32_b_offset] "I" (offsetof(arm_gemm::Requantize32, b_offset)), [offsetof_Requantize32_c_offset] "I" (offsetof(arm_gemm::Requantize32, c_offset)), [offsetof_Requantize32_maxval] "I" (offsetof(arm_gemm::Requantize32, maxval)), [offsetof_Requantize32_minval] "I" (offsetof(arm_gemm::Requantize32, minval)), [outptrs] "r" (outptrs), [qp] "r" (&qp)
    : "cc", "memory", "p0", "p1", "p2", "x9", "x10", "x11", "x19", "x20", "x21", "x22", "x23", "x24", "x25", "x26", "x27", "x28", "z0", "z1", "z2", "z3", "z4", "z5", "z6", "z7", "z8", "z9", "z10", "z11", "z12", "z16", "z17", "z18", "z19", "z20", "z21", "z22", "z23", "z24", "z25", "z26", "z27", "z28", "z29", "z30", "z31"
  );
}

}  // namespace depthwise
}  // namespace arm_conv

#endif  // defined(ARM_COMPUTE_ENABLE_SVE)
